@@ -30,7 +30,7 @@ read_credentials <- function(file = default_credentials_file()) {
 
 #' @rdname read_credentials
 #' @export
-use_credentials <- function(profile = "default", file = default_credentials_file()) {
+use_credentials <- function(profile = Sys.getenv("AWS_PROFILE", "default"), file = default_credentials_file()) {
     if (inherits(file, "aws_credentials")) {
         x <- file
     } else {
@@ -68,7 +68,7 @@ parse_credentials <- function(char) {
     make_named_vec <- function(x) {
         elem <- strsplit(x, "[ ]?=[ ]?")
         out <- lapply(elem, `[`, 2)
-        names(out) <- toupper(sapply(elem, `[`, 1))
+        names(out) <- trimws(toupper(sapply(elem, `[`, 1)))
         out
     }
 
